"""
Module containing utility functions
"""

import datetime as dt
import re
import string
from typing import Any

from andar.check_utils import check_expected_fields
from andar.field_conf import FieldConf


def get_template_fields_names(path_template: str) -> list[str]:
    """
    Get fields names from path template string

    Params:
        path_template: Path template that follows string.Formatter() syntax.

    Returns:
        Template fields names.
    """
    parsed_field_tuples = list(string.Formatter().parse(path_template))
    template_fields_names = [name for (text, name, spec, conv) in parsed_field_tuples if name is not None]
    return template_fields_names


def assign_groupname_pattern_dict(pattern_dict: dict[str:str]) -> dict[str:str]:
    """
    Assign a group name to each regex pattern present in the given dictionary

    Params:
        pattern_dict: A dictionary of regex patterns, where each key will be used as group name. It does not
                  check if the pattern already have a group name assign.

    Returns:
        A dictionary where the patterns have been assigned a group name.
    """
    named_pattern_dict = {}
    for field, pattern in pattern_dict.items():
        named_pattern_dict[field] = f"(?P<{field}>{pattern})"
    return named_pattern_dict


def compile_path_regex(template: str, fields: dict[str:FieldConf], ds: str = "/") -> re.Pattern:
    """
    Compile a regex pattern using a path template and a FieldConf dictionary

    Example:
        ```python
        filename_template = "{prefix}_{name}.{extension}"
        fields = {
            "prefix": FieldConf(pattern=r"[0-9]{4}"),
            "name": FieldConf(pattern=r"[a-zA-Z0-9]+"),
            "extension": FieldConf(pattern=r"json")
        }
        filename = "0001_example.json"
        compiled_regex = compile_path_regex(filename_template, fields)
        compiled_regex.match(filename).groupdict()
        ```

    Params:
        template: A path template that follows `PathModel.template` syntax.
        fields: A dictionary where each key represent a field of the template and each value represent its corresponding
                `FieldConf()`.
        ds: Directory separator string.

    Returns:
        A compiled regex pattern.
    """
    template_field_names = get_template_fields_names(template)
    field_names = list(fields.keys())
    check_expected_fields(template_field_names, field_names)

    invalid_fields = [n for n in field_names if n.find("__") >= 0]

    if len(invalid_fields) > 0:
        raise ValueError(f"Fields cannot contains double underscore: {invalid_fields}")

    template = template.replace(r".", r"\.")
    pattern_dict = {}
    for field_name, field_conf in fields.items():
        field_pattern = field_conf.pattern
        if field_conf.is_optional:
            field_pattern = f"{field_pattern}|"
            # Allow optional directory separator for this field: "/" -> "/?", by updating path_template
            field_name_dir_sep = "{" + field_name + "}" + ds + "{"
            optional_field_name_dir_sep = "{" + field_name + "}" + ds + "?{"
            template = template.replace(field_name_dir_sep, optional_field_name_dir_sep)
        pattern_dict[field_name] = field_pattern

    # Deduplicate repeated fields of pattern_dict:
    # for example the template "{base_path}/{asset_name}/{asset_name}_{suffix}"
    # will become "{base_path}/{asset_name__0}/{asset_name__1}_{suffix}"
    # and the dict {"base_path": r"\w+", "asset_name": r"\w+", "suffix": r"\d+"}
    # will become {"base_path": r"\w+", "asset_name__0": r"\w+", "asset_name__1": r"\w+", "suffix": r"\d+"}
    unique_fields = list(set(template_field_names))
    new_pattern_dict = {}
    new_template = template
    for field_name in unique_fields:
        field_count = len([f for f in template_field_names if f == field_name])
        if field_count == 1:
            new_pattern_dict[field_name] = pattern_dict[field_name]
            continue
        for idx in range(field_count):
            new_field_name = field_name + f"__{idx}"
            new_pattern_dict[new_field_name] = pattern_dict[field_name]
            new_template = new_template.replace("{" + field_name + "}", "{" + new_field_name + "}", 1)

    has_duplicates = pattern_dict != new_pattern_dict
    if has_duplicates:
        pattern_dict = new_pattern_dict
        template = new_template

    # Build full pattern string
    named_pattern_dict = assign_groupname_pattern_dict(pattern_dict)
    path_pattern = template.format(**named_pattern_dict)
    path_pattern = f"^{path_pattern}$"  # match the full string
    compiled_pattern = re.compile(path_pattern)
    return compiled_pattern


def fusion_deduplicated_fields(parsed_field_dict: dict[str, str]) -> dict[str, str]:
    """
    Fusion deduplicated fields in a parsed field dict

    Example:
        ```python
        deduplicated_field_dict = {
            "name": "data",
            "version__0": "v1",
            "version__1": "v1",
            "ext": "csv",
        }
        fusioned_field_dict = fusion_deduplicated_fields(deduplicated_field_dict)
        # {
        #     "name": "data",
        #     "version": "v1",
        #     "ext": "csv",
        # }
        ```

    Params:
        parsed_field_dict: A dictionary where each key represent a field name. A deduplicated field will have a
                           suffix `__{id}`.

    Returns:
        A dictionary where deduplicated fields are fusioned in a single field with the suffix removed.
    """

    deduplicated_regex = re.compile(r"^(?P<name>.+?)__(?P<id>\d)$")

    deduplicated_fields_dict = {}

    for field_name in parsed_field_dict:
        match = deduplicated_regex.match(field_name)
        if match is None:
            continue
        match_dict = match.groupdict()
        original_field_name = match_dict["name"]
        if original_field_name not in deduplicated_fields_dict:
            deduplicated_fields_dict[original_field_name] = []
        deduplicated_fields_dict[original_field_name].append(field_name)

    # Fusion deduplicated fields:
    # it will raise an error if the deduplicate fields have multiples values
    # for example this parsed dict will raise an error because asset_name__0 and asset_name__1 should be equal:
    # {"base_path": "folder", "asset_name__0": "my_asset", "asset_name__1": "other_asset", "suffix": "001"}
    # if the deduplicated fields are coherent they will be fusion and renamed to its original name:
    # for example {"base_path": "folder", "asset_name__0": "my_asset", "asset_name__1": "my_asset", "suffix": "001"}
    # to {"base_path": "folder", "asset_name": "my_asset", "suffix": "001"}
    for original_field_name, deduplicated_list in deduplicated_fields_dict.items():
        parsed_field_values = [parsed_field_dict.pop(f) for f in deduplicated_list]
        unique_parsed_field_values = list(set(parsed_field_values))
        are_duplicated_unique = len(unique_parsed_field_values) == 1
        if not are_duplicated_unique:
            raise ValueError(
                f"More than one value was found for repeated field '{original_field_name}': {parsed_field_values}"
            )
        parsed_field_dict[original_field_name] = unique_parsed_field_values[0]
    return parsed_field_dict


def prepare_fields_values(fields_values_dict: dict[str, Any], fields_conf: dict[str, FieldConf]) -> dict[str, str]:
    """
    Prepare fields values for this path

    Params:
        fields_values_dict: Dictionary of fields values
        fields_conf: Dictionary of fields configuration (i.e.class FieldConf)

    Returns:
        A dictionary of fields where the values were converted to strings.
    """
    new_fields_values_dict = {}
    for field_name, field_value in fields_values_dict.items():
        if field_name not in fields_conf:
            print(f"skipping field '{field_name}'")
            continue
        field_conf = fields_conf[field_name]

        if field_value is None and field_conf.is_optional:
            new_fields_values_dict[field_name] = ""
            continue

        if field_conf.date_format is not None:
            new_field_value = field_value.strftime(field_conf.date_format)
        elif field_conf.datetime_format is not None:
            new_field_value = field_value.strftime(field_conf.datetime_format)
        elif field_conf.var_to_str is not None:
            new_field_value = field_conf.var_to_str(field_value)
        else:
            new_field_value = str(field_value)

        field_pattern = f"^{field_conf.pattern}$"  # Exact pattern
        result = re.match(field_pattern, new_field_value)
        if result is None:
            raise ValueError(
                f"Invalid field '{field_name}' value: '{new_field_value}'. It does not match pattern: "
                f"'{field_conf.pattern}'"
            )
        new_fields_values_dict[field_name] = new_field_value
    return new_fields_values_dict


def process_parsed_fields_values(fields_conf: dict[str, FieldConf], parsed_fields: dict[str, str]) -> dict[str, Any]:
    """
    Process fields values dictionary obtained from parsing a file path

    Params:
        parsed_fields: A dictionary of parsed fields values in string format.

    Returns:
        A processed dictionary of fields with converted values depending on each FieldConf definition.
    """
    new_parsed_fields = parsed_fields.copy()

    for field_name, field_value in new_parsed_fields.items():
        if field_name not in fields_conf:
            raise ValueError(f"Unknown field '{field_name}'. Valid fields are: {fields_conf.keys()}")
        field_conf = fields_conf[field_name]

        field_pattern = f"^{field_conf.pattern}$"  # Exact pattern
        if field_conf.is_optional:
            field_pattern = f"^{field_conf.pattern}|$"
        result = re.match(field_pattern, field_value)
        if result is None:
            raise ValueError(
                f"Invalid field '{field_name}' value: '{field_value}'. It does not match pattern: "
                f"'{field_conf.pattern}'"
            )

        if field_conf.date_format is not None:
            new_field_value = dt.datetime.strptime(field_value, field_conf.date_format).date()
        elif field_conf.datetime_format is not None:
            new_field_value = dt.datetime.strptime(field_value, field_conf.datetime_format)
        elif field_conf.str_to_var is not None:
            new_field_value = field_conf.str_to_var(field_value)
        else:
            new_field_value = str(field_value)

        if new_field_value == "" and field_conf.is_optional:
            new_field_value = None

        new_parsed_fields[field_name] = new_field_value
    return new_parsed_fields



import os.path
import re
from typing import Any, Self

from andar.check_utils import check_expected_fields, check_parent_path_template
from andar.field_conf import FieldConf
from andar.parser_utils import (
    compile_path_regex,
    fusion_deduplicated_fields,
    get_template_fields_names,
    prepare_fields_values,
    process_parsed_fields_values,
)


class PathModel:
    """
    PathModel allows to define, build and parse templated file paths

    It defines a path via a template and its fields. Once instantiated, it allows to build new paths
    and to parse path strings to recover individual fields.

    Attributes:
        template: A template string
        fields: A dictionary of FieldConfs
    """

    template: str
    fields: dict[str, FieldConf]
    default_field: FieldConf
    parent_template: str
    description: str
    compiled_regex: re.Pattern
    _dir_sep: str = "/"

    def __init__(
        self,
        template: str,
        *,
        fields: dict[str, FieldConf] | None = None,
        default_field: FieldConf | None = None,
        parent_template: str | None = None,
        description: str | None = None,
    ):
        self.template = template
        self._fields = fields
        self._parent_template = parent_template
        self._default_field = default_field
        self._description = description

        if parent_template is None:
            parent_template = os.path.dirname(self.template)
        check_parent_path_template(template, parent_template)
        self.parent_template = parent_template

        if default_field is None:
            default_field = FieldConf()
        self.default_field = default_field

        new_fields = {}
        template_field_names = get_template_fields_names(template)
        for field_name in template_field_names:
            new_fields[field_name] = default_field

        if fields is not None:
            new_fields.update(fields)
        self.fields = new_fields

        new_field_names = list(new_fields.keys())
        check_expected_fields(template_field_names, new_field_names)

        if description is None:
            description = ""
        self.description = description

        self.compiled_regex = compile_path_regex(
            template=self.template,
            fields=self.fields,
            ds=self._dir_sep,
        )

    def __repr__(self):
        ident = "  "
        formatted_fields = "\n".join([f"{ident * 2}'{k}': {v}," for k, v in self.fields.items()])
        formatted_args = [f"{ident}template='{self.template}',"]
        if self._fields is not None:
            formatted_args.append(f"{ident}fields={{\n{formatted_fields}\n  }},")
        if self._default_field is not None:
            formatted_args.append(f"{ident}default_field={self.default_field},")
        if self._parent_template is not None:
            formatted_args.append(f"{ident}parent_template='{self.parent_template}',")
        if self._description is not None:
            formatted_args.append(f"{ident}description='{self.description}',")
        formatted_args_str = "\n".join(formatted_args)
        repr = f"PathModel(\n{formatted_args_str}\n)"
        return repr

    def replace(self, copy_description: bool = False, **kwargs: Any) -> Self:
        """
        Creates a copy of the current object replacing attributes with the given keyword arguments

        Params:
            **kwargs: Attributes to be replaced, same arguments as used for PathModel instantiation

        Returns:
            A PathModel instance
        """
        default_parent_template = None  # default value when a new template is given, if not it reuse a previous one
        if "template" not in kwargs:
            kwargs["template"] = self.template
            default_parent_template = self._parent_template
        if "parent_template" not in kwargs:
            kwargs["parent_template"] = default_parent_template  # reset to None when a new template is set
        if "fields" not in kwargs:
            kwargs["fields"] = self._fields
        if "default_field" not in kwargs:
            kwargs["default_field"] = self._default_field
        default_description = None
        if copy_description:
            default_description = self._description
        if "description" not in kwargs:
            kwargs["description"] = default_description

        return self.__class__(**kwargs)

    def update(self, **kwargs: Any) -> Self:
        """
        Creates a copy of the current object updating attributes with the given keyword arguments

        Params:
            **kwargs: Attributes to be updated, same arguments as used for PathModel instantiation.
                      Fields set to `None` will be reset to default_field, if no longer present in the template,
                      they will be removed instead.

        Returns:
            A PathModel instance
        """
        fields = self.fields.copy()
        if "fields" in kwargs:
            fields.update(kwargs["fields"])
            fields_names = list(fields.keys())
            [fields.pop(n) for n in fields_names if fields[n] is None]  # remove fields set to None from input args
            kwargs["fields"] = fields
        return self.replace(**kwargs)

    def __call__(self, **kwargs) -> Self:
        if not kwargs:
            return self
        return self.replace(**kwargs)

    def parse_path(self, file_path: str, raise_error: bool = False) -> dict[str:Any]:
        """
        Parse a file path

        Params:
            file_path: String to be parsed.
            raise_error: Whether to raise an exception if the file path is not valid. By default, it returns None.

        Returns:
            Dictionary where each key represent a field of the template and each value is the corresponding parsed
            string (or converted object)
        """

        match = self.compiled_regex.match(file_path)
        if not match:
            if raise_error:
                raise ValueError(f"Invalid path '{file_path}', expected pattern: '{self.compiled_regex.pattern}'")
            return None
        parsed_fields_dict = match.groupdict()
        parsed_fields_dict = fusion_deduplicated_fields(parsed_fields_dict)

        processed_fields = process_parsed_fields_values(self.fields, parsed_fields_dict)
        return processed_fields

    @classmethod
    def _get_path(
        cls,
        template: str,
        fields_conf: dict[str, FieldConf],
        fields_values_dict: dict[str, Any],
    ) -> str:
        """
        Generate path using input parameters

        Params:
            template: A template that follows string.Formatter() syntax.
            fields_conf: Dictionary of fields configurations, where keys are field names and values are FieldConf
                         instances.
            fields_values_dict: Input parameters dict that maps template fields to values.

        Returns:
            A path string.
        """
        fields_values_dict = fields_values_dict.copy()
        template_field_names = get_template_fields_names(template)
        missing_field_names = [field for field in template_field_names if field not in fields_values_dict]

        for field_name in missing_field_names:
            is_optional = fields_conf[field_name].is_optional
            if is_optional:
                fields_values_dict[field_name] = None

        given_field_names = list(fields_values_dict.keys())
        check_expected_fields(template_field_names, given_field_names)
        fields_values_dict = prepare_fields_values(fields_values_dict, fields_conf)
        built_path = template.format(**fields_values_dict)

        # Clean path of duplicated slashes using normpath() when optional fields are present
        normalized_path = os.path.normpath(built_path)
        if built_path.startswith("./"):
            # for preserving bijection of first field when field value is a dot `"."`
            normalized_path = "./" + normalized_path

        return normalized_path

    def get_path(self, **kwargs: Any) -> str:
        """
        Generate path using input parameters

        Params:
            **kwargs: Input parameters that maps template fields to values.

        Returns:
            A path string.
        """
        return self._get_path(template=self.template, fields_conf=self.fields, fields_values_dict=kwargs)

    def get_parent_path(self, **kwargs: Any) -> str:
        """
        Generate parent path using input parameters

        Params:
            **kwargs: Input parameters that maps template fields to values. They are used in the order of
                      parent_template, if the last argument(s) are omitted, the parent_template will be dynamically
                      updated to a shorter version. If an argument in the middle is omitted, and it is not optional,
                      an error will be raised.

        Returns:
            Parent path string.
        """

        # remove all fields not present in parent template
        parent_fields = self.fields.copy()
        fields_names = get_template_fields_names(self.template)
        parent_fields_names = get_template_fields_names(self.parent_template)

        # Drop fields corresponding to filename
        for field in fields_names:
            if field not in parent_fields_names:
                parent_fields.pop(field)

        # Make dynamic the parent path creation, so the last fields can be omitted
        # for example "{a}/{b}/{c}/{d}/" or "{a}/{b}/{c}/" or "{a}/{b}/"
        new_parent_template = self.parent_template
        dynamic_fields_names = []
        missing_kwarg_name = None
        for parent_field_name in parent_fields_names:
            is_optional = parent_fields[parent_field_name].is_optional
            if parent_field_name not in kwargs and not is_optional:
                # keep known (left) part of parent_template and drop the rest
                missing_kwarg_name = parent_field_name
                new_parent_template = new_parent_template.split("{" + parent_field_name + "}", 1)[0]
                break
            dynamic_fields_names.append(parent_field_name)

        # Drop fields corresponding to last arguments that were skipped
        parent_fields_names = list(parent_fields.keys())
        for field_name in parent_fields_names:
            if field_name not in dynamic_fields_names:
                parent_fields.pop(field_name)

        # Check for unnecessary extra args
        extra_kwargs = {}
        for field_name in kwargs:
            if field_name not in parent_fields:
                extra_kwargs[field_name] = kwargs[field_name]
        if missing_kwarg_name and extra_kwargs:
            raise ValueError(
                f"Unexpected extra kwargs: {extra_kwargs}, after updating parent template "
                f"to '{new_parent_template}' because of missing kwarg: '{missing_kwarg_name}'"
            )

        return self._get_path(
            template=new_parent_template,
            fields_conf=parent_fields,
            fields_values_dict=kwargs,
        )

    def assert_path_bijection(self, test_path: str):
        """
        Assert path bijection

        It tries to recover the same initial input after processing once with parse_file_path and get_path

        Params:
            test_path: Path string to be tested
        """
        parsed_fields = self.parse_path(test_path, raise_error=True)
        result_test_path = self.get_path(**parsed_fields)
        assert test_path == result_test_path, f"{test_path} != {result_test_path}"

    def assert_fields_bijection(self, test_fields: dict[str, Any]):
        """
        Assert fields bijection

        It tries to recover the same initial input after processing once with get_path and parse_file_path
        This method is the preferred way of checking if the PathModel was well-defined.

        Params:
            test_fields: Dictionary of fields to be tested
        """
        test_path = self.get_path(**test_fields)
        result_parsed_fields = self.parse_path(test_path, raise_error=True)
        assert test_fields == result_parsed_fields, f"{test_fields} != {result_parsed_fields}"



import datetime as dt
import unittest

from andar.examples import data_mesh_pm, date_archived_pm


class ExamplesTests(unittest.TestCase):
    def test_date_archived_pm(self):
        report_date = dt.date(year=2025, month=6, day=1)
        run_datetime = dt.datetime(year=2025, month=12, day=1, hour=12, minute=15, second=30)
        result_path = date_archived_pm.get_path(
            base_path="company_name/new_project",
            subfolder="reports",
            date_path=report_date,
            date_prefix=report_date,
            name="clients",
            datetime_suffix=run_datetime,
            ext="csv",
        )
        expected_path = "/company_name/new_project/reports/2025/06/01/2025-06-01_clients_20251201_121530.csv"
        self.assertEqual(expected_path, result_path)

        date_archived_pm.assert_path_bijection(expected_path)

    def test_data_mesh_pm(self):
        campaing_performane_date = dt.date(year=2025, month=12, day=1)
        result_path = data_mesh_pm.get_path(
            domain="marketing",
            layer="mart",
            product="campaing_performane",
            aggregation="monthly",
            date=campaing_performane_date,
            ext="csv",
        )
        expected_path = "/marketing/mart/campaing_performane/monthly/20251201_campaing_performane.csv"
        self.assertEqual(expected_path, result_path)

        data_mesh_pm.assert_path_bijection(expected_path)



import datetime as dt
import unittest

from andar import FieldConf, PathModel, SafePatterns


class PathModelTests(unittest.TestCase):
    def test_simple_template(self):
        path_builder = PathModel(template="{base_folder}/{intermediate_folder}/{base_name}_{suffix}.{extension}")
        result_path = path_builder.get_path(
            base_folder="parent_folder",
            intermediate_folder="other_folder",
            base_name="my_data",
            suffix="2000-01-01",
            extension="csv",
        )
        expected_path = "parent_folder/other_folder/my_data_2000-01-01.csv"
        self.assertEqual(expected_path, result_path)

        path_builder.parse_path(expected_path)

        wrong_path = "folder/my_data.csv"
        with self.assertRaises(ValueError) as cm:
            path_builder.parse_path(wrong_path, raise_error=True)
        expected_error_msg = f"Invalid path '{wrong_path}', expected pattern"
        self.assertIn(expected_error_msg, str(cm.exception))

        result_parent_path = path_builder.get_parent_path(
            base_folder="parent_folder",
            intermediate_folder="other_folder",
        )
        expected_parent_path = "parent_folder/other_folder"
        self.assertEqual(expected_parent_path, result_parent_path)

        path_builder = PathModel(
            template="{base_folder}/{intermediate_folder}/{base_name}_{suffix}.{extension}",
            parent_template="{base_folder}",
        )
        result_parent_path = path_builder.get_parent_path(base_folder="parent_folder")
        expected_parent_path = "parent_folder"
        self.assertEqual(expected_parent_path, result_parent_path)

    def test_custom_fields(self):
        new_path_builder = PathModel(
            template="{base_path}/{intermediate_folder}/{base_name}_{suffix}.{extension}",
            fields={
                "base_path": FieldConf(pattern=SafePatterns.DIRPATH),
                "intermediate_folder": FieldConf(pattern=r"\d{4}-\d{2}-\d{2}", date_format="%Y-%m-%d"),
                "base_name": FieldConf(
                    pattern=SafePatterns.NAME,
                    var_to_str=str.upper,
                    str_to_var=str.lower,
                ),
                "suffix": FieldConf(
                    pattern=r"\d{4}-\d{2}-\d{2}_\d{6}",
                    datetime_format="%Y-%m-%d_%H%M%S",
                ),
                "extension": FieldConf(pattern=r"[a-z]+"),
            },
        )

        custom_datetime = dt.datetime.fromisoformat("2025-02-01 12:34:56")
        custom_date = custom_datetime.date()

        test_path = new_path_builder.get_path(
            base_path="/parent/folder",
            intermediate_folder=custom_date,
            base_name="my_data",
            suffix=custom_datetime,
            extension="csv",
        )

        expected_test_path = "/parent/folder/2025-02-01/MY_DATA_2025-02-01_123456.csv"
        self.assertEqual(expected_test_path, test_path)

        test_parent_path = new_path_builder.get_parent_path(
            base_path="/parent/folder",
            intermediate_folder=custom_date,
        )
        expected_test_parent_path = "/parent/folder/2025-02-01"
        self.assertEqual(expected_test_parent_path, test_parent_path)

        new_path_builder.assert_path_bijection(expected_test_path)
        input_fields = {
            "base_path": "/parent/folder",
            "intermediate_folder": custom_date,
            "base_name": "my_data",
            "suffix": custom_datetime,
            "extension": "csv",
        }
        new_path_builder.assert_fields_bijection(input_fields)

        # test relative base_path `.`:
        test_relative_path = "./2025-02-01/MY_DATA_2025-02-01_123456.csv"
        new_path_builder.assert_path_bijection(test_relative_path)
        input_fields = {
            "base_path": ".",
            "intermediate_folder": custom_date,
            "base_name": "my_data",
            "suffix": custom_datetime,
            "extension": "csv",
        }
        new_path_builder.assert_fields_bijection(input_fields)

    def test_optional_fields(self):
        optional_path_builder = PathModel(
            template="{base_path}/{env}/{intermediate_folder}/{base_name}_{suffix}.{extension}",
            fields={
                "base_path": FieldConf(pattern=SafePatterns.DIRPATH),
                "env": FieldConf(pattern=r"dev|preprod|prod|local", is_optional=True),
                "suffix": FieldConf(pattern=SafePatterns.FIELD, is_optional=True),
            },
        )
        # Test get_path
        test_path = optional_path_builder.get_path(
            env="dev",
            base_path="/parent_folder",
            intermediate_folder="sub_folder",
            base_name="my_data",
            suffix="suffix",
            extension="csv",
        )
        expected_test_path = "/parent_folder/dev/sub_folder/my_data_suffix.csv"
        self.assertEqual(expected_test_path, test_path)

        # Test raise error when omitting a mandatory field
        with self.assertRaises(ValueError) as context:
            optional_path_builder.get_path(
                env="dev",
                # base_path is mandatory, and it should not be omitted,
                intermediate_folder="sub_folder",
                # base_name is mandatory, and it should not be omitted,
                # suffix is optional and can be omitted
                extension="csv",
            )
        expected_key_error_msg = "Missing fields: ['base_path', 'base_name'] they are required in expected field list"
        self.assertIn(expected_key_error_msg, str(context.exception))

        # Test get_parent_path omitting a field in the parent_path
        test_parent_path = optional_path_builder.get_parent_path(
            # env="dev" is optional and can be omitted
            base_path="/parent_folder",
            intermediate_folder="sub_folder",
        )
        expected_test_parent_path = "/parent_folder/sub_folder"
        self.assertEqual(expected_test_parent_path, test_parent_path)

        # Test get_parent_path omitting a field in the name
        test_path = optional_path_builder.get_path(
            env="dev",
            base_path="/parent_folder",
            intermediate_folder="sub_folder",
            base_name="my_data",
            # suffix="custom_suffix" is optional and can be omitted
            extension="csv",
        )
        expected_test_path = "/parent_folder/dev/sub_folder/my_data_.csv"
        self.assertEqual(expected_test_path, test_path)

        optional_path_builder.assert_path_bijection(expected_test_path)
        input_fields = {
            "base_path": "/parent_folder",
            "env": "dev",
            "intermediate_folder": "sub_folder",
            "base_name": "my_data",
            "suffix": None,
            "extension": "csv",
        }
        optional_path_builder.assert_fields_bijection(input_fields)

    def test_path_builder_converters(self):
        custom_datetime = dt.datetime.fromisoformat("2025-02-01 12:34:56")
        custom_date = custom_datetime.date()
        optional_path_builder = PathModel(
            template="{base_path}/{env}/{intermediate_folder}/{base_name}_{suffix}.{extension}",
            fields={
                "env": FieldConf(pattern=r"dev|preprod|prod|local", is_optional=True),
                "base_path": FieldConf(pattern=SafePatterns.DIRPATH),
                "intermediate_folder": FieldConf(pattern=r"\d{4}-\d{2}-\d{2}", date_format="%Y-%m-%d"),
                "base_name": FieldConf(
                    pattern=SafePatterns.NAME,
                    var_to_str=str.upper,
                    str_to_var=str.lower,
                ),
                "suffix": FieldConf(
                    pattern=r"\d{4}-\d{2}-\d{2}_\d{6}",
                    datetime_format="%Y-%m-%d_%H%M%S",
                ),
                "extension": FieldConf(pattern=r"[a-z]+"),
            },
        )

        test_path = optional_path_builder.get_path(
            env="dev",
            base_path="/parent/folder",
            intermediate_folder=custom_date,
            base_name="my_data",
            suffix=custom_datetime,
            extension="csv",
        )
        expected_test_path = "/parent/folder/dev/2025-02-01/MY_DATA_2025-02-01_123456.csv"
        self.assertEqual(expected_test_path, test_path)

        test_path = optional_path_builder.get_path(
            # env="dev" is optional and can be omitted
            base_path="/parent/folder",
            intermediate_folder=custom_date,
            base_name="my_data",
            suffix=custom_datetime,
            extension="csv",
        )
        expected_test_path = "/parent/folder/2025-02-01/MY_DATA_2025-02-01_123456.csv"
        self.assertEqual(expected_test_path, test_path)

        test_parent_path = optional_path_builder.get_parent_path(
            env="dev",
            base_path="/parent/folder",
            intermediate_folder=custom_date,
        )
        expected_test_parent_path = "/parent/folder/dev/2025-02-01"
        self.assertEqual(expected_test_parent_path, test_parent_path)

        test_parent_path = optional_path_builder.get_parent_path(
            # env="dev" is optional and can be omitted
            base_path="/parent/folder",
            intermediate_folder=custom_date,
        )
        expected_test_parent_path = "/parent/folder/2025-02-01"
        self.assertEqual(expected_test_parent_path, test_parent_path)

        optional_path_builder.assert_path_bijection(test_path)
        input_fields = {
            "base_path": "/parent/folder",
            "env": None,
            "intermediate_folder": custom_date,
            "base_name": "my_data",
            "suffix": custom_datetime,
            "extension": "csv",
        }
        optional_path_builder.assert_fields_bijection(input_fields)

    def test_unknown_field(self):
        with self.assertRaises(ValueError) as context:
            PathModel(
                template="{folder_a}/{base_name}.{extension}",
                fields={"unknown_field": FieldConf(pattern=SafePatterns.FIELD)},
            )
        expected_error_msg = "Invalid fields: ['unknown_field'] they do not exist in expected field list"
        self.assertIn(expected_error_msg, str(context.exception))

    def test_repeated_fields(self):
        path_builder = PathModel(
            template="{folder_a}/{version}/{base_name}_{version}.{extension}",
        )
        test_path = path_builder.get_path(
            folder_a="aaa",
            base_name="filename",
            version="v1",
            extension="txt",
        )
        expected_test_path = "aaa/v1/filename_v1.txt"
        self.assertEqual(expected_test_path, test_path)

        fields_dict = path_builder.parse_path(expected_test_path)
        expected_fields = {
            "folder_a": "aaa",
            "base_name": "filename",
            "extension": "txt",
            "version": "v1",
        }
        self.assertEqual(expected_fields, fields_dict)

    def test_path_builder_replace(self):
        path_builder = PathModel(
            template="{folder_a}/{folder_b}/{folder_c}/{base_name}{suffix}.{extension}",
            fields={
                "folder_a": FieldConf(pattern=SafePatterns.FIELD, is_optional=True),
            },
        )

        new_path_builder = path_builder.replace(
            fields={
                "folder_b": FieldConf(pattern=SafePatterns.FIELD, is_optional=True),
            }
        )
        test_parent_path = new_path_builder.get_parent_path(folder_a="aaa", folder_c="ccc")
        expected_test_parent_path = "aaa/ccc"
        self.assertEqual(expected_test_parent_path, test_parent_path)

        new_path_builder = path_builder.replace(
            template="{folder_b}/{folder_a}/{folder_c}/{base_name}{suffix}.{extension}",
        )
        test_path = new_path_builder.get_path(
            folder_a="aaa",
            folder_b="bbb",
            folder_c="ccc",
            base_name="filename",
            suffix="123",
            extension="txt",
        )
        expected_test_path = "bbb/aaa/ccc/filename123.txt"
        self.assertEqual(expected_test_path, test_path)

    def test_path_builder_update(self):
        path_builder = PathModel(
            template="{folder_a}/{folder_b}/{folder_c}/{base_name}{suffix}.{extension}",
            fields={
                "folder_b": FieldConf(pattern=SafePatterns.FIELD, is_optional=True),
            },
        )

        new_path_builder = path_builder.update(
            fields={
                "folder_c": FieldConf(pattern=SafePatterns.FIELD, is_optional=True),
            }
        )
        test_parent_path = new_path_builder.get_parent_path(folder_a="aaa")  # now folder_b and folder_c are optional
        expected_test_parent_path = "aaa"
        self.assertEqual(expected_test_parent_path, test_parent_path)

    def test_bijection_errors(self):
        path_builder = PathModel(
            template="{a}_{b}_{c}_{d}.{e}",
            fields={
                "c": FieldConf(is_optional=True),
                "d": FieldConf(is_optional=True),
            },
        )
        misleading_path = "a_A_b_B.e"
        path_builder.assert_path_bijection(misleading_path)  # assert_path_bijection does NOT detect this error

        invalid_fields = {"a": "a_A", "b": "b_B", "e": "e"}
        with self.assertRaises(AssertionError) as cm:
            path_builder.assert_fields_bijection(invalid_fields)  # assert_fields_bijection does DETECT the error
        expected_msg = "{'a': 'a_A', 'b': 'b_B', 'e': 'e'} != {'a': 'a', 'b': 'A', 'c': 'b', 'd': 'B__', 'e': 'e'}"
        self.assertIn(expected_msg, str(cm.exception))

    def test_dynamic_parent_path_creation(self):
        path_builder = PathModel("/{a}/{b}/{c}/{d}/{name}")
        test_parent_path = path_builder.get_parent_path(a="aaa", b="bbb", c="ccc")
        expected_parent_path = "/aaa/bbb/ccc"
        self.assertEqual(expected_parent_path, test_parent_path)

        test_parent_path = path_builder.get_parent_path(a="aaa", b="bbb")
        expected_parent_path = "/aaa/bbb"
        self.assertEqual(expected_parent_path, test_parent_path)

        test_parent_path = path_builder.get_parent_path(a="aaa")
        expected_parent_path = "/aaa"
        self.assertEqual(expected_parent_path, test_parent_path)

        test_parent_path = path_builder.get_parent_path()
        expected_parent_path = "/"
        self.assertEqual(expected_parent_path, test_parent_path)

        with self.assertRaises(ValueError) as context:
            path_builder.get_parent_path(a="aaa", c="ccc")
        expected_key_error_msg = (
            "Unexpected extra kwargs: {'c': 'ccc'}, after updating parent template to '/{a}/' "
            "because of missing kwarg: 'b'"
        )
        self.assertIn(expected_key_error_msg, str(context.exception))


if __name__ == "__main__":
    unittest.main()



from collections.abc import Callable
from dataclasses import dataclass, replace
from typing import Any, Self


class SafePatterns:
    """
    Non-greedy patterns

    Attributes:
        FIELD: Alphanumeric characters
        NAME: Alphanumeric characters plus name separators: `-_.`
        DIRPATH: Alphanumeric characters plus name separators: `-_.` plus directory separator: `/`
        EXTENSION: Alphanumeric characters plus dots for sub extensions as in `tar.gz`
    """

    FIELD = r"[a-zA-Z0-9]+?"  # without separator characters
    NAME = r"[-_.a-zA-Z0-9]+?"  # include name separators: -_.
    DIRPATH = r"[-_.a-zA-Z0-9/]+?"  # include name separators: -_. and directory separator: /
    EXTENSION = r"[.a-zA-Z0-9]+?"  # include dots for sub extensions as 'tar.gz'


@dataclass
class FieldConf:
    """
    FieldConf allows to configure how a path field is validated, parsed and processed

    Params:
        pattern: Regex pattern used to validate the input of get_path and get_parent_path, and also to get fields
                 using PathModel.parse_path(). By default, it used SafePatterns.NAME.
        description: String used as documentation for the FieldConf.
        date_format: date_format and datetime_format are used in get_path and get_parent_path for accepting either a
                        string or a datetime object that can be formated to an input string for a the given template.
                        They are also used for validating the parsed field when using parse_path and for defining how
                        to cast the string (to date or to datetime, depending on the argument that was used).
        datetime_format: See description of date_format parameter.
        is_optional: allows to omit a field during the path generation (get_path and get_parent_path)
                        or to skip field during the path parsing. IMPORTANT! in order to work properly, pattern must
                        be constrained, otherwise the PathModel class may have an unexpected behaviour.
                        For example use '[0-9]{4}' and avoid using '*' or '+' when using is_optional=True.
    """

    pattern: str = SafePatterns.NAME
    description: str | None = None
    date_format: str | None = None
    datetime_format: str | None = None
    is_optional: bool | None = None
    str_to_var: Callable | None = None
    var_to_str: Callable | None = None

    def __post_init__(self):
        has_date_converter = self.date_format is not None
        has_datetime_converter = self.datetime_format is not None
        has_custom_converter = self.var_to_str is not None or self.str_to_var is not None
        active_converters_num = sum([has_date_converter, has_datetime_converter, has_custom_converter])
        if active_converters_num > 1:
            raise ValueError(f"Maximum one field converter is allowed, but {active_converters_num} were found")

    def __repr__(self):
        formatted_fields = []
        for k, v in vars(self).items():
            if v is None:
                continue
            v_str = f"'{v}'" if isinstance(v, str) else v.__qualname__
            field_str = f"{k}={v_str}"
            formatted_fields.append(field_str)
        formatted_fields_str = ", ".join(formatted_fields)
        repr = f"FieldConf({formatted_fields_str})"
        return repr

    def replace(self, **kwargs: Any) -> Self:
        """
        Creates a copy of the current object replacing attributes with the given keyword arguments

        Params:
            **kwargs: Attributes (parameters) to be replaced

        Returns:
            A copy of the object with the attributes replaced
        """
        field_conf = replace(self, **kwargs)
        return field_conf
